"""scratchpad for testing and implementing pwspyqc-functions."""

import numpy as np
import scipy
import tqdm
import xarray as xr


def _indicator_correlation(
    a_dataset, b_dataset, prob, exclude_nan=True, min_valid_overlap=None
):
    """Calculate indicator correlation two datasets.

    Parameters
    ----------
    a_dataset : np.array
        First data vector
    b_dataset : np.array
        Second data vector
    perc : float
        Percentile threshold
    exclude_nan : bool
        Default True, exculdes pairs where a least one value is NaN
    min_valid_overlap : int
        Minimum number of required overlapping data for calculating the
        indicator correlation

    Returns
    -------
    indicator correlation values [np.array]
    """
    if len(a_dataset.shape) != 1:
        msg = "`a_dataset` has to be a 1D numpy.ndarray"
        raise ValueError(msg)
    if a_dataset.shape != b_dataset.shape:
        msg = "`a_dataset` and `b_dataset` have to have the same shape"
        raise ValueError(msg)

    if (a_dataset < 0).any() or (b_dataset < 0).any():
        msg = "input arrays must not contain negative values"
        raise ValueError(msg)

    a_dataset = np.copy(a_dataset)
    b_dataset = np.copy(b_dataset)

    both_not_nan = ~np.isnan(a_dataset) & ~np.isnan(b_dataset)
    if exclude_nan:
        a_dataset = a_dataset[both_not_nan]
        b_dataset = b_dataset[both_not_nan]

    if min_valid_overlap is not None:
        if sum(both_not_nan) < min_valid_overlap:
            return np.nan
    else:  # noqa: PLR5501
        if sum(both_not_nan) == 0:
            msg = "No overlapping data. Define `min_valid_overlap` to return NaN in such cases."  # noqa: E501
            raise ValueError(msg)

    # Get index at quantile threshold `prob`
    ix = int(a_dataset.shape[0] * prob)

    # Set values below quantile threshold `prob` to 0
    # and above to 1
    a_sort = np.sort(a_dataset)
    b_sort = np.sort(b_dataset)
    a_dataset[a_dataset < a_sort[ix]] = 0
    b_dataset[b_dataset < b_sort[ix]] = 0
    a_dataset[a_dataset > 0] = 1
    b_dataset[b_dataset > 0] = 1

    # Calculate correlation of 0 and 1 time series
    cc = np.corrcoef(a_dataset, b_dataset)[0, 1]
    return cc  # noqa: RET504


def indicator_distance_matrix(
    da_a,
    da_b,
    max_distance=50000,  # this is in meters, assuming the projection units are also meters # noqa: E501
    prob=0.99,
    exclude_nan=True,
    min_valid_overlap=None,
):
    """Calculate indicator correlation and distance between reference and test stations.

    Parameters
    ----------
    da_a : xr.DataArray
        First data vector, has to be in the OpensSense data format standards [1]
        with rainfall as variable
    da_b : xr.DataArray
        Second data vector, has to be in the OpensSense data format standards [1]
        with rainfall as variable
    max_distance : int
        Maximum distance in meters for which the indicator correlation is returned
    prob : float
        Percentile threshold for indicator correlation
    exclude_nan : bool
        Default True, exculdes pairs where a least one value is NaN
    min_valid_overlap : int
        Minimum number of overlapping data for calculating the indicator correlation

    Literature
    ----------
    [1] Fencl M, Nebuloni R, C. M. Andersson J et al. Data formats and standards for
    opportunistic rainfall sensors [version 2; peer review: 2 approved].
    Open Res Europe 2024, 3:169 (https://doi.org/10.12688/openreseurope.16068.2)

    Returns
    -------
    Indicator correlation and distance matrices as tuple [xr.DataArray, xr.DataArray]

    """
    xy_a = list(zip(da_a.x.data, da_a.y.data))
    xy_b = list(zip(da_b.x.data, da_b.y.data))
    dist_mtx = scipy.spatial.distance.cdist(xy_a, xy_b, metric="euclidean")
    indcorr_mtx = np.full_like(dist_mtx, np.nan)
    # list_corr = []
    # list_dist = []
    for i in tqdm.tqdm(range(len(xy_a))):
        for j in range(len(xy_b)):
            # check if distance between stations is less than max_distance
            if dist_mtx[i, j] < max_distance:
                ts_a = da_a.isel(id=i)
                ts_b = da_b.isel(id=j)
                ts_b = ts_b.reindex({"time": ts_a.time})

                indcorr_mtx[i, j] = _indicator_correlation(
                    ts_a.data,
                    ts_b.data,
                    prob=prob,
                    exclude_nan=exclude_nan,
                    min_valid_overlap=min_valid_overlap,
                )
    da_dist_mtx = xr.DataArray(
        data=dist_mtx,
        dims=("id", "id_neighbor"),
        coords={
            "id": ("id", da_a.id.data),
            "id_neighbor": ("id_neighbor", da_b.id.data),
        },
    )
    da_indcorr_mtx = xr.DataArray(
        data=indcorr_mtx,
        dims=("id", "id_neighbor"),
        coords={
            "id": ("id", da_a.id.data),
            "id_neighbor": ("id_neighbor", da_b.id.data),
        },
    )
    return da_dist_mtx, da_indcorr_mtx


def ic_filter(
    indicator_correlation_matrix_ref,
    distance_correlation_matrix_ref,
    indicator_correlation_matrix,
    distance_matrix,
    max_distance=20e3,
    bin_size=1e3,
    quantile_bin_ref=0.1,
    quantile_bin_pws=0.5,
    threshold=0.01,
):
    """Apply indicator correlation filter [1].

    This function applies a modified version of the indicator correlation filter from
    Bárdossy el al. (2021). [1]

    Parameters
    ----------
    indicator_correlation_matrix_ref: xr.DataArray
        Indicator correlation matrix between reference stations (REF)
    distance_correlation_matrix_ref: xr.DataArray
        Distance matrix between reference stations (REF)
    indicator_correlation_matrix: xr.DataArray
        Indicator correlations matrix between REF and PWS
    distance_matrix: xr.DataArray
        Distance matrix between REF and PWS
    max_distance: int or float
        Range in meters for which the indicator correlation is evaluated
    bin_size: int or float
        Bin size in meters. This bin size is used to group data for the
        quantile calculation
    quantile_bin_ref: float
        Quantile for acceptance level based on reference data indicator correlation
    quantile_bin_pws: float
        Quantile of PWS data indicator correlation that has to be
        above `quantile_bin_ref` + `threshold`
    threshold: float
        Indicator correlation threshold below `quantile_bin_ref` where PWS are
        still accepted

    Literature
    ----------
    [1] Bárdossy, A., Seidel, J., and El Hachem, A.: The use of personal weather station
    observations to improve precipitation estimation and interpolation,
    Hydrol. Earth Syst. Sci., 25, 583-601,
    https://doi.org/10.5194/hess-25-583-2021, 2021.

    Returns
    -------
    distance matrix between PWS and reference stations [xr.DataArray]
    indicator correlation matrix between PWS and reference stations [xr.DataArray]
    boolean if station got accepted [xr.DataArray]
    indicator correlation score [xr.DataArray]
    """
    bins = np.arange(0, max_distance, bin_size)

    # quantile parameter not too low, otherwise the line becomes to wiggly - depends on data  # noqa: E501
    binned_indcorr_ref = (
        indicator_correlation_matrix_ref.groupby_bins(
            distance_correlation_matrix_ref, bins=bins
        )
        .quantile(quantile_bin_ref)
        .bfill(dim="group_bins")
    )

    # Function for Rank Sum Weights
    # Calculates weights according to length to data set
    def rsw(m):
        alphas = []  # Leere Liste
        for i in range(1, m + 1):  # Iteration über m Alternativen
            alpha = (m + 1.0 - i) / sum(range(1, m + 1))
            alphas.append(alpha)
        return alphas

    pws_indcorr_good_list = []
    pws_indcorr_score_list = []

    # iterates over REF (id)
    for pws_id in indicator_correlation_matrix["id_neighbor"].values:  # noqa: PD011
        binned_indcorr_pws = (
            indicator_correlation_matrix.sel(id_neighbor=pws_id)
            .groupby_bins(distance_matrix.sel(id_neighbor=pws_id), bins=bins)
            .quantile(quantile_bin_pws, skipna=True)
        )

        indcorr_good = binned_indcorr_pws + threshold > binned_indcorr_ref

        # Bool Information if PWS passed Indicator Correlation Test
        pws_indcorr_good_list.append(indcorr_good.any())

        # Valid bins for normed weights
        valid_bins = np.isfinite(binned_indcorr_pws.values)
        rank_sum_weights = rsw(len(indcorr_good))
        normed_weights = sum(valid_bins * np.array(rank_sum_weights))

        score = sum(indcorr_good.values * np.array(rank_sum_weights)) / normed_weights  # noqa: PD011
        pws_indcorr_score_list.append(score)

    result = indicator_correlation_matrix.to_dataset(name="indcorr")
    result["dist"] = distance_matrix
    result["indcorr_good"] = ("id_neighbor", pws_indcorr_good_list)
    result["indcorr_score"] = ("id_neighbor", pws_indcorr_score_list)

    return result



"""A collection of functions for flagging problematic time steps."""

from __future__ import annotations

import numpy as np
import numpy.typing as npt
import pandas as pd
import xarray as xr


def fz_filter(
    ds_pws: npt.NDArray[np.float64],
    nint: npt.NDArray[np.float64],
    n_stat: npt.NDArray[np.float64],
) -> npt.NDArray[np.float64]:
    """Faulty Zeros Filter.

    This function applies the FZ filter from the R package PWSQC.
    The flag 1 means, a faulty zero has been detected. The flag -1
    means that no flagging was done because evaluation cannot be
    performed for the first `nint` values.

    Note that this code here is derived from the Python translation,
    done by Niek van Andel, of the original R code from Lotte de Vos.
    The Python code stems from here https://github.com/NiekvanAndel/QC_radar.
    Also note that the correctness of the Python code has not been
    verified and not all feature of the R implementation might be there.

    Parameters
    ----------
    pws_data
        The rainfall time series of the PWS that should be flagged
    reference
        The rainfall time series of the reference, which can be e.g.
        the median of neighboring PWS data.
    nint : optional
        The number of subsequent data points which have to be zero, while
        the reference has values larger than zero, to set the flag for
        this data point to 1.

    Returns
    -------
    npt.NDArray
        time series of flags
    """
    pws_data = ds_pws.rainfall
    nbrs_not_nan = ds_pws.nbrs_not_nan
    reference = ds_pws.reference

    # find first rainfall observation in each time series
    first_non_nan_index = ds_pws["rainfall"].notnull().argmax(dim="time")  # noqa: PD004

    # Create a mask that is True up to the first valid
    # index for each station, False afterward
    mask = xr.DataArray(
        np.arange(ds_pws.sizes["time"]), dims="time"
    ) < first_non_nan_index.broadcast_like(ds_pws["rainfall"])

    # initialize arrays
    sensor_array = np.zeros_like(pws_data)
    ref_array = np.zeros_like(pws_data)
    fz_array = np.zeros_like(pws_data)

    # Wet timestep at each station
    sensor_array[np.where(pws_data > 0)] = 1

    # Dry timestep at each station
    sensor_array[np.where(pws_data == 0)] = 0

    # Wet timesteps of the reference
    ref_array[np.where(reference > 0)] = 1

    for i in np.arange(len(pws_data.id.data)):
        for j in np.arange(len(pws_data.time.data)):
            if j < nint:
                fz_array[i, j] = -1
            elif sensor_array[i, j] > 0:
                fz_array[i, j] = 0
            elif fz_array[i, j - 1] == 1:
                fz_array[i, j] = 1
            elif (np.sum(sensor_array[i, j - nint : j + 1]) > 0) or (
                np.sum(ref_array[i, j - nint : j + 1]) < nint + 1
            ):
                fz_array[i, j] = 0
            else:
                fz_array[i, j] = 1

    fz_array = fz_array.astype(int)
    fz_flag = xr.where(nbrs_not_nan < n_stat, -1, fz_array)

    # add to dataset
    ds_pws["fz_flag"] = fz_flag

    # set fz_flag to -1 up to the first valid rainfall observation
    ds_pws["fz_flag"] = ds_pws["fz_flag"].where(~mask, -1)

    # check if last nint timesteps are NaN in rolling window
    nan_in_last_nint = (
        ds_pws["rainfall"].rolling(time=nint, center=True).construct("window_dim")
    )
    all_nan_in_window = nan_in_last_nint.isnull().all(dim="window_dim")

    # Apply the mask to set fz_flag to -1 where the condition is met
    ds_pws["fz_flag"] = ds_pws["fz_flag"].where(~all_nan_in_window, -1)

    return ds_pws


def hi_filter(
    ds_pws: npt.NDArray[np.float64],
    hi_thres_a: npt.NDArray[np.float64],
    hi_thres_b: npt.NDArray[np.float64],
    nint: npt.NDArray[np.float64],
    n_stat=npt.NDArray[np.float64],
) -> npt.NDArray[np.float64]:
    """High Influx filter.

    This function applies the HI filter from the R package PWSQC,
    flagging unrealistically high rainfall amounts.

    The Python code has been translated from the original R code,
    to be found here: https://github.com/LottedeVos/PWSQC/tree/master/R.

    The function returns an array with zeros, ones or -1 per time step
    and station.
    The flag 0 means that no high influx has been detected.
    The flag 1 means that high influx has been detected.
    The flag -1 means that no flagging was done because not enough
    neighbouring stations are reporting rainfall to make a reliable
    evaluation.

    Parameters
    ----------
    pws_data
        The rainfall time series of the PWS that should be flagged
    nbrs_not_nan
        Number of neighbouring stations reporting rainfall
    reference
        The rainfall time series of the reference, which can be e.g.
        the median of neighboring stations
    hi_thres_a
        threshold for median rainfall of neighbouring stations [mm]
    hi_thres_b
        upper rainfall limit [mm]
    n_stat
        threshold for number of neighbours reporting rainfall

    Returns
    -------
    npt.NDArray
        time series of flags
    """
    # find first rainfall observation in each time series
    first_non_nan_index = ds_pws["rainfall"].notnull().argmax(dim="time")  # noqa: PD004

    # Create a mask that is True up to the first
    # valid index for each station, False afterward
    mask = xr.DataArray(
        np.arange(ds_pws.sizes["time"]), dims="time"
    ) < first_non_nan_index.broadcast_like(ds_pws["rainfall"])

    condition1 = (ds_pws.reference < hi_thres_a) & (ds_pws.rainfall > hi_thres_b)
    condition2 = (ds_pws.reference >= hi_thres_a) & (
        ds_pws.rainfall > ds_pws.reference * hi_thres_b / hi_thres_a
    )

    hi_array = (condition1 | condition2).astype(int)

    hi_flag = xr.where(ds_pws.nbrs_not_nan < n_stat, -1, hi_array)

    # add to dataset
    ds_pws["hi_flag"] = hi_flag

    # set hi_flag to -1 up to the first valid rainfall observation
    ds_pws["hi_flag"] = ds_pws["hi_flag"].where(~mask, -1)

    # check if last nint timesteps are NaN in rolling window
    nan_in_last_nint = (
        ds_pws["rainfall"].rolling(time=nint, center=True).construct("window_dim")
    )
    all_nan_in_window = nan_in_last_nint.isnull().all(dim="window_dim")

    # Apply the mask to set hi_flag to -1 where the condition is met
    ds_pws["hi_flag"] = ds_pws["hi_flag"].where(~all_nan_in_window, -1)

    return ds_pws


def so_filter_one_station(da_station, da_neighbors, evaluation_period, mmatch):
    """Support function to Station Outlier filter.

    Parameters
    ----------
    da_station
        rainfall time series of evaluated station.
    da_neighbors
        rainfall time series of neighboring stations.
    evaluation_period
        length of (rolling) window for correlation calculation
        [timesteps]
    mmatch
        threshold for number of matching rainy intervals in
        evaluation period [timesteps]

    Returns
    -------
    npt.NDArray
        number of neighbors with enough wet time steps
    """
    # rolling pearson correlation
    s_station = da_station.to_series()
    s_neighbors = da_neighbors.to_series()
    corr = s_station.rolling(evaluation_period, min_periods=1).corr(s_neighbors)
    ds = xr.Dataset.from_dataframe(pd.DataFrame({"corr": corr}))

    # create dataframe of neighboring stations
    df_nbrs = da_neighbors.to_dataframe()
    df_nbrs = df_nbrs["rainfall"].unstack("id")  # noqa: PD010

    # boolean arrays - True if a rainy time step, False if 0 or NaN.
    rainy_timestep_at_nbrs = df_nbrs > 0

    # rolling sum of number of rainy timesteps in
    # last evaluation_period period, per neighbor.
    wet_timesteps_last_evaluation_period_period = rainy_timestep_at_nbrs.rolling(
        evaluation_period, min_periods=1
    ).sum()

    # per time step and neighbor, does the nbr have more than
    # mmatch wet time steps in the last evaluation_period period? (true/false)
    enough_matches_per_nbr = wet_timesteps_last_evaluation_period_period > mmatch

    # summing how many neighbors that have enough matches per time step
    nr_nbrs_with_enough_matches = enough_matches_per_nbr.sum(axis=1)

    ds["matches"] = xr.DataArray.from_series(nr_nbrs_with_enough_matches)

    return ds


def so_filter(
    ds_pws: npt.NDArray[np.float64],
    distance_matrix: npt.NDArray[np.float64],
    evaluation_period: npt.NDArray[np.float64],
    mmatch: npt.NDArray[np.float64],
    gamma: npt.NDArray[np.float64],
    n_stat=npt.NDArray[np.float64],
    max_distance=npt.NDArray[np.float64],
) -> npt.NDArray[np.float64]:
    """Station Outlier filter.

    This function applies the SO filter from the R package PWSQC,
    flagging nonsensical rainfall measurements for a specific location.

    The Python code has been translated from the original R code,
    to be found here: https://github.com/LottedeVos/PWSQC/tree/master/R.

    In its original implementation, any interval with at least `mrain`
    intervals of nonzero rainfall measurements is evaluated.
    In this implementation, only a fixed rolling window of `evaluation_period`
    intervals is evaluated.

    The function returns an array with zeros, ones or -1 per time step
    and station.
    The flag 0 means that no station outlier has been detected.
    The flag 1 means that a station outlier has been detected.
    The flag -1 means that no flagging was done because not enough
    neighbouring stations are reporting rainfall to make a reliable
    evaluation or that the previous evaluation_period time steps was dry.

    Parameters
    ----------
    ds_pws
        xarray data set
    nbrs_not_nan
        Number of neighbouring stations reporting rainfall
    evaluation_period
        length of (rolling) window for correlation calculation
        [timesteps]
    mmatch
        threshold for number of matching rainy intervals in
        evaluation period [timesteps]
    gamma
        threshold for rolling median pearson correlation [-]
    n_stat
        threshold for number of neighbours reporting rainfall
    max_distance
        considered range around each station [m]

    Returns
    -------
    npt.NDArray
        Time series of flags.
    """
    # For each station (ID), get the index of the first non-NaN rainfall value
    first_non_nan_index = ds_pws["rainfall"].notnull().argmax(dim="time")  # noqa: PD004

    for i in range(len(ds_pws.id)):
        ds_station = ds_pws.isel(id=i)
        pws_id = ds_station.id.to_numpy()

        # picking stations within max_distnance, excluding itself,
        # for the whole duration of the time series
        neighbor_ids = distance_matrix.id.data[
            (distance_matrix.sel(id=pws_id) < max_distance)
            & (distance_matrix.sel(id=pws_id) > 0)
        ]

        # create data set for neighbors
        ds_neighbors = ds_pws.sel(id=neighbor_ids)

        # if there are no observations in the time series, filter
        # cannot be applied to the whole time series
        # or if there are not enough stations nearby,
        # filter cannot be applied to the whole time series
        if ds_pws.rainfall.sel(id=pws_id).isnull().all() or (
            len(neighbor_ids) < n_stat
        ):
            ds_pws.so_flag[i, :] = -1
            ds_pws.median_corr_nbrs[i, :] = -1
            continue

        # run so-filter
        ds_so_filter = so_filter_one_station(
            ds_station.rainfall, ds_neighbors.rainfall, evaluation_period, mmatch
        )

        median_correlation = ds_so_filter.corr.median(dim="id", skipna=True)
        ds_pws.median_corr_nbrs[i] = median_correlation

        so_array = (median_correlation < gamma).astype(int)

        # filter can not be applied if less than n_stat neighbors have enough matches
        ds_pws.so_flag[i] = xr.where(ds_so_filter.matches < n_stat, -1, so_array)

        # Set so_flag to -1 up to first valid index
        first_valid_time = first_non_nan_index[i].item()
        ds_pws["so_flag"][i, :first_valid_time] = -1

        # disregard warm up period
        ds_pws.so_flag[
            i, first_valid_time : (first_valid_time + evaluation_period)
        ] = -1

    return ds_pws



"""Skript with functions for plots for the peak removal filter."""

# import packages
import matplotlib.pyplot as plt
import numpy as np
from matplotlib.lines import Line2D
from matplotlib.patches import Circle


def plot_station_neighbors(
    a_dataset,
    b_dataset,
    station,
    aa_closest_neighbors,
    ab_closest_neighbors,
    max_distance,
    zoom=True,
):
    """Plot the neighbor stations of a selected station.

    Parameters
    ----------
    a_dataset : xr.DataArray
        Dataset, following the OpenSense data format standards.
    b_dataset : xr.DataArray
        Dataset, following the OpenSense data format standards.
    station : str
        Name/number of the station.
    aa_closest_neighbors : xr.Dataset
        Dataset containing the closest neighbors of a_dataset with a_dataset as
        neighbor, the distances of the neighbors and the ids of the neighbors.
    ab_closest_neighbors : xr.Dataset
        Dataset containing the closest neighbors of a_dataset with b_dataset as
        neighbor, the distances of the neighbors and the ids of the neighbors.
    max_distance : float
        The allowed distance of neighbors has to be smaller than max_distance.
        The unites are the units used for the projected coordinates x and y in the
        two datasets.

        Has to be the same as in closest_neighbors.
    zoom : bool
        If True, the plot is zoomed in on the station. If False,
        the plot is not zoomed in.

    Returns
    -------
    None
        Plot of the station and its neighbors.
    """
    # get the x and y coordinates of the station
    x_station = a_dataset.sel(id=station).x.to_numpy()
    y_station = a_dataset.sel(id=station).y.to_numpy()

    # get the x and y coordinates of all stations and get the neighbors of the station
    x = a_dataset.x.to_numpy()
    y = a_dataset.y.to_numpy()
    aa_neighbor = aa_closest_neighbors.sel(id=station).neighbor_id.to_numpy()
    if b_dataset is not None:
        x_ref = b_dataset.x.to_numpy()
        y_ref = b_dataset.y.to_numpy()
        ab_neigbors = ab_closest_neighbors.sel(id=station).neighbor_id.to_numpy()

    s = 10 if zoom else 2

    # plot all pws stations and reference stations
    fig, ax = plt.subplots(dpi=150)
    plt.scatter(x=x, y=y, s=s, color="blue", alpha=0.5)
    if b_dataset is not None:
        plt.scatter(x=x_ref, y=y_ref, s=s, color="black", alpha=0.5)

    for neighbor in aa_neighbor:
        if neighbor is None:
            continue
        plt.scatter(
            a_dataset.sel(id=neighbor).x.to_numpy(),
            a_dataset.sel(id=neighbor).y.to_numpy(),
            s=s,
            color="lime",
            alpha=0.5,
        )
    if b_dataset is not None:
        for neighbor_ref in ab_neigbors:
            if neighbor_ref is None:
                continue
            plt.scatter(
                b_dataset.sel(id=neighbor_ref).x.to_numpy(),
                b_dataset.sel(id=neighbor_ref).y.to_numpy(),
                s=s,
                color="magenta",
                alpha=0.5,
            )

    # highlight the station
    plt.scatter(x_station, y_station, s=s, color="red", label=f"pws station {station}")

    # plot search radius
    kreis = Circle(
        (x_station, y_station),
        radius=max_distance,
        color="black",
        linewidth=0.5,
        fill=False,
    )
    plt.gca().add_patch(kreis)
    plt.axis("equal")

    # create custom legend
    if b_dataset is not None:
        custom_legend = [
            Line2D(
                [0],
                [0],
                marker="o",
                color="w",
                markerfacecolor="blue",
                markersize=4,
                label="pws network",
            ),
            Line2D(
                [0],
                [0],
                marker="o",
                color="w",
                markerfacecolor="black",
                markersize=4,
                label="reference network",
            ),
            Line2D(
                [0],
                [0],
                marker="o",
                color="w",
                markerfacecolor="lime",
                markersize=4,
                label="pws neighbors",
            ),
            Line2D(
                [0],
                [0],
                marker="o",
                color="w",
                markerfacecolor="magenta",
                markersize=4,
                label="reference neighbors",
            ),
            Line2D(
                [0],
                [0],
                marker="o",
                color="w",
                markerfacecolor="red",
                markersize=4,
                label=f"pws station {station}",
            ),
            Line2D(
                [0],
                [0],
                marker="o",
                color="w",
                markeredgecolor="black",
                markeredgewidth=0.5,
                label=f"search radius: {max_distance/1000} km",
            ),
        ]
    else:
        custom_legend = [
            Line2D(
                [0],
                [0],
                marker="o",
                color="w",
                markerfacecolor="blue",
                markersize=4,
                label="pws network",
            ),
            Line2D(
                [0],
                [0],
                marker="o",
                color="w",
                markerfacecolor="lime",
                markersize=4,
                label="pws neighbors",
            ),
            Line2D(
                [0],
                [0],
                marker="o",
                color="w",
                markerfacecolor="red",
                markersize=4,
                label=f"pws station {station}",
            ),
            Line2D(
                [0],
                [0],
                marker="o",
                color="w",
                markeredgecolor="black",
                markeredgewidth=0.5,
                label=f"search radius: {max_distance/1000} km",
            ),
        ]

    # set zoom scope of the plot
    if zoom:
        plt.xlim(x_station - max_distance * 3, x_station + max_distance * 3)
        plt.ylim(y_station - max_distance * 3, y_station + max_distance * 3)

    name_plot = f"Neighbors of station {station}"
    plt.xlabel("longitude")
    plt.ylabel("latitude")
    plt.title(name_plot)
    plt.legend(handles=custom_legend)

    plt.show()
    plt.close()
    return fig, ax


def plot_peak(
    dataset,
    data_corr,
    station,
    quantile,
    peak_num,
    seq_start_lst,
    time_peak_lst,
    seq_end_lst,
    seq_len_lst,
    data_is_corrected=False,
    zoom_out=15,
):
    """Plot the rainfall of the station with the (corrected) peaks.

    Parameters
    ----------
    dataset : xr.DataArray
        Dataset, following the OpenSense data format standards.
    data_corr : xr.Dataset
        Dataset, following the OpenSense data format standards with the corrected peaks
        and nan sequences.

        If not available, set data_corr to None and data_is_corrected to False.
    station : str
        Name/number of the station.
    quantile : float
        Quantile for the peak determination.
    peak_num : int
        Number of the peak to plot.
    seq_start_lst : list
        List of start times of the nan sequences.
    time_peak_lst : list
        List of times of the peaks.
    seq_end_lst : list
        List of end times of the nan sequences.
    seq_len_lst : list
        List of lengths of the nan sequences.
    data_is_corrected : bool
        Set True, if the dataset is corrected and given as input for data_corr.
    zoom_out : int
        Factor to additionally zoom out the plot. 1 = 5 min.

    Returns
    -------
    None
        Plot of the rainfall of the station with the (corrected) peaks.
    """
    # set timegap and timegap_plt
    timegap = int(
        (dataset.time.to_numpy()[1] - dataset.time.to_numpy()[0])
        / np.timedelta64(1, "m")
    )
    timedelta = np.timedelta64(timegap, "m")
    timegap_plt_before = timedelta * seq_len_lst[peak_num] + zoom_out * timedelta
    timegap_plt_after = timedelta * seq_len_lst[peak_num] + zoom_out * timedelta

    # create time series needed for the plot
    if (
        seq_start_lst[peak_num] - timegap_plt_before
        not in dataset.sel(id=station).time.to_numpy()
    ):
        start = dataset.sel(id=station).time.to_numpy()[0]
    else:
        start = seq_start_lst[peak_num] - timegap_plt_before
    if (
        seq_end_lst[peak_num] + timegap_plt_after
        not in dataset.sel(id=station).time.to_numpy()
    ):
        end = dataset.sel(id=station).time.to_numpy()[-1]
    else:
        end = time_peak_lst[peak_num] + timegap_plt_after

    x = np.arange(start, end + timedelta, timedelta)
    x_nan_seq = np.arange(
        seq_start_lst[peak_num], seq_end_lst[peak_num] + timedelta * 2, timedelta
    )  # includes the peak
    x_nan_seq_others = [
        np.arange(seq_start, seq_end + timedelta * 2, timedelta)
        for seq_start, seq_end in zip(seq_start_lst, seq_end_lst)
    ]  # include the peak
    x_peak = np.datetime64(time_peak_lst[peak_num])

    # get the values of the time series
    y_peak_orig = dataset.sel(id=station, time=x_peak).rainfall.to_numpy()
    if data_is_corrected:
        y = data_corr.sel(id=station, time=x).rainfall.to_numpy()
        y_nan_seq = data_corr.sel(id=station, time=x_nan_seq).rainfall.to_numpy()
        y_nan_seq_others = [
            data_corr.sel(id=station, time=x_nan_seq_other).rainfall.to_numpy()
            for x_nan_seq_other in x_nan_seq_others
        ]
        y_peak = data_corr.sel(id=station, time=x_peak).rainfall.to_numpy()
    else:
        y = dataset.sel(id=station, time=x).rainfall.to_numpy()
        y_nan_seq = dataset.sel(id=station, time=x_nan_seq).rainfall.to_numpy()
        y_nan_seq_others = [
            dataset.sel(id=station, time=x_nan_seq_other).rainfall.to_numpy()
            for x_nan_seq_other in x_nan_seq_others
        ]
        y_peak = dataset.sel(id=station, time=x_peak).rainfall.to_numpy()

    fig, ax = plt.subplots()
    if data_is_corrected:
        name_plot_1 = f"Corrected rainfall of station {station} with peak no."
        name_plot_2 = f"{peak_num + 1} out of {len(time_peak_lst)}"
        name_plot = f"{name_plot_1} {name_plot_2}"

    else:
        name_plot_1 = f"Rainfall of station {station} with peak no."
        name_plot_2 = f"{peak_num + 1} out of {len(time_peak_lst)}"
        name_plot = f"{name_plot_1} {name_plot_2}"

    # plot the time series
    ax.stem(x, y, markerfmt="o", linefmt="blue", basefmt=" ").markerline.set_markersize(
        2.5
    )

    # plot the leading nan sequence/the distributed peak values for the nan sequence
    if y_peak == y_peak_orig:
        ax.stem(
            x_nan_seq, y_nan_seq, markerfmt="o", linefmt="red", basefmt=" "
        ).markerline.set_markersize(2.5)
    else:
        ax.stem(
            x_nan_seq, y_nan_seq, markerfmt="o", linefmt="lime", basefmt=" "
        ).markerline.set_markersize(2.5)

    # plot the peak/the new value
    if y_peak == y_peak_orig:
        ax.stem(
            x_peak, y_peak, markerfmt="o", linefmt="red", basefmt=" "
        ).markerline.set_markersize(2.5)
    else:
        ax.stem(
            x_peak, y_peak, markerfmt="o", linefmt="lime", basefmt=" "
        ).markerline.set_markersize(2.5)

    # mark other peaks/remaining or corrected peaks
    for time_peak_other, x_nan_seq_other, y_nan_seq_other in zip(
        time_peak_lst, x_nan_seq_others, y_nan_seq_others
    ):
        if time_peak_other in x:
            if time_peak_other == x_peak:  # skip the above selected peak
                continue
            y_peak_other = dataset.sel(
                id=station, time=time_peak_other
            ).rainfall.to_numpy()
            if data_is_corrected:
                y_peak_other_corr = data_corr.sel(
                    id=station, time=time_peak_other
                ).rainfall.to_numpy()
                if (
                    y_peak_other == y_peak_other_corr
                ):  # check if the peak is corrected --> green, if not --> orange
                    ax.stem(
                        x_nan_seq_other,
                        y_nan_seq_other,
                        markerfmt="o",
                        linefmt="orange",
                        basefmt=" ",
                    ).markerline.set_markersize(2.5)
                else:
                    ax.stem(
                        x_nan_seq_other,
                        y_nan_seq_other,
                        markerfmt="o",
                        linefmt="green",
                        basefmt=" ",
                    ).markerline.set_markersize(2.5)
            else:
                # plot other peaks
                ax.stem(
                    time_peak_other,
                    y_peak_other,
                    markerfmt="o",
                    linefmt="orange",
                    basefmt=" ",
                ).markerline.set_markersize(2.5)
        else:
            continue

    # create custom legend
    if data_is_corrected:
        if y_peak == y_peak_orig:
            custom_legend = [
                Line2D(
                    [0],
                    [0],
                    marker="|",
                    color="red",
                    markersize=8,
                    markeredgewidth=1.5,
                    linestyle="",
                    label="not corrected peak",
                ),
                Line2D(
                    [0],
                    [0],
                    marker="|",
                    color="blue",
                    markersize=8,
                    markeredgewidth=1.5,
                    linestyle="",
                    label="legit values",
                ),
                Line2D(
                    [0],
                    [0],
                    marker="|",
                    color="orange",
                    markersize=8,
                    markeredgewidth=1.5,
                    linestyle="",
                    label="remaining peaks",
                ),
                Line2D(
                    [0],
                    [0],
                    marker="|",
                    color="green",
                    markersize=8,
                    markeredgewidth=1.5,
                    linestyle="",
                    label="other corrected peaks",
                ),
                Line2D([0], [0], linestyle="--", color="black", label="threshold"),
            ]
        else:
            custom_legend = [
                Line2D(
                    [0],
                    [0],
                    marker="|",
                    color="lime",
                    markersize=8,
                    markeredgewidth=1.5,
                    linestyle="",
                    label="corrected peak",
                ),
                Line2D(
                    [0],
                    [0],
                    marker="|",
                    color="blue",
                    markersize=8,
                    markeredgewidth=1.5,
                    linestyle="",
                    label="legit values",
                ),
                Line2D(
                    [0],
                    [0],
                    marker="|",
                    color="orange",
                    markersize=8,
                    markeredgewidth=1.5,
                    linestyle="",
                    label="remaining peaks",
                ),
                Line2D(
                    [0],
                    [0],
                    marker="|",
                    color="green",
                    markersize=8,
                    markeredgewidth=1.5,
                    linestyle="",
                    label="other corrected peaks",
                ),
                Line2D([0], [0], linestyle="--", color="black", label="threshold"),
            ]
    else:
        custom_legend = [
            Line2D(
                [0],
                [0],
                marker="|",
                color="red",
                markersize=8,
                markeredgewidth=1.5,
                linestyle="",
                label="peak",
            ),
            Line2D(
                [0],
                [0],
                marker="|",
                color="blue",
                markersize=8,
                markeredgewidth=1.5,
                linestyle="",
                label="legit values",
            ),
            Line2D(
                [0],
                [0],
                marker="|",
                color="orange",
                markersize=8,
                markeredgewidth=1.5,
                linestyle="",
                label="other peaks",
            ),
            Line2D([0], [0], linestyle="--", color="black", label="threshold"),
        ]

    ax.grid(True)

    # plot horizontal line at the threshold
    threshold = np.nanquantile(dataset.sel(id=station).rainfall.to_numpy(), quantile)
    plt.axhline(y=threshold, color="black", linestyle="--")

    # set xlim and x ticks
    plt.xlim(x[0], x[-1])
    plt.xticks(rotation=45)

    plt.title(name_plot)
    plt.xlabel("time")
    plt.ylabel("rainfall [mm]")
    plt.legend(handles=custom_legend, loc="upper left")
    plt.show()
    plt.close()
    return fig, ax
